#include <motor_control.h>
#include <multi_motor_control.h>
#include <iostream>
#include <thread>
#include <chrono>

#define delay(s) std::this_thread::sleep_for(std::chrono::milliseconds((int32_t)(s*1000)))

int main()
{
    #define status_packet1 0xb1
    #define status_packet2 0xb2
    #define status_packet3 0xb3

    // port_name: Windows: "COM*"  Linux: "/dev/ttyUSB*"
    // id: Executor ID
    // Baudrate: The baud rate can only be 921600
    string port_name = "COM6";
    uint32_t Baudrate = 921600;
    uint8_t id = 0x00;
    MotorControl motor(port_name, Baudrate, id);

    // CURRENT_MODE
    // Feedback_cycle1(ms): Feedback period of the first status packet，
    //                      Including voltage, pwm, current, speed
    // Feedback_cycle2(ms)：Feedback period of the second status packet，
    //                      Including the position of the motor and the position of the actuator
    // Feedback_cycle3(ms)：Feedback period of the third status packet，
    //                      Including MOS tube, motor, chip temperature and warning information, 
    //                      error information, current operation mode
    // cur: current(mA)
    uint8_t Feedback_cycle1 = 10;
    uint8_t Feedback_cycle2 = 10;
    uint8_t Feedback_cycle3 = 10;
    int16_t cur = 1000;

    motor.reset();
    motor.mode_selection(Mode::CURRENT_MODE, Feedback_cycle1, Feedback_cycle2, Feedback_cycle3);
    motor.write_current(cur);

    cout << endl << "current mode" << endl;
    printf("target_current: %d\n", cur);
    delay(2);

    MotorStatus status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("current: %d\n", status.current_value);
    

    // PROFILE_VELOCITY_MODE
    // cur(mA): limit current
    // vel(rpm): target velocity of motor
    cur = 10000;
    int16_t vel = 500;

    motor.reset();
    motor.mode_selection(Mode::PROFILE_VELOCITY_MODE, Feedback_cycle1, Feedback_cycle2, Feedback_cycle3);
    motor.write_profile_velocity(cur, vel);

    cout << endl << "profile velocity mode" << endl;
    printf("target_velocity: %d\n", vel);
    delay(2);
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("velocity: %d\n", status.velocity_value);

    motor.write_profile_velocity(cur, 0);
    delay(1);
    

    // PROFILE_POSITION_MODE
    // cur(mA): limit current 
    // vel(rpm): the limit velocity of motor
    // pos(count): target position, With a 16-bit encoder, the actuator counts 65536 per revolution.
    cur = 10000;
    vel = 1000;
    int32_t pos = 20000;

    motor.reset();
    delay(0.6);
    motor.mode_selection(Mode::PROFILE_POSITION_MODE, Feedback_cycle1, Feedback_cycle2, Feedback_cycle3);
    delay(0.6);
    motor.write_profile_position(cur, vel, 0);
    delay(2);
    motor.write_profile_position(cur, vel, pos);

    cout << endl << "profile position mode" << endl;
    printf("target_position: %d\n", pos);
    delay(5);
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("position: %d\n", status.actuator_position_value);

    motor.write_profile_position(cur, vel, 0);
    delay(3);

    // POSITION_MODE
    motor.reset();
    motor.mode_selection(Mode::POSITION_MODE, Feedback_cycle1, Feedback_cycle2, Feedback_cycle3);
    motor.write_position(cur, vel, pos);

    cout << endl << "position mode" << endl;
    printf("target_position: %d\n", pos);
    delay(5);
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("position: %d\n", status.actuator_position_value);

    motor.write_position(cur, vel, 0);

    // MIT_MODE(Refer to the tutorial on how to use MIT mode)
    cur = 1000;
    vel = 0;
    pos = 0;
    float kp = 0;
    float kd = 0;
    motor.reset();
    motor.mode_selection(Mode::MIT_MODE, Feedback_cycle1, Feedback_cycle2, Feedback_cycle3);

    // control current
    motor.write_mit(cur, vel, pos, kp, kd);

    cout << endl << "MIT mode" << endl;
    printf("target_current: %d\n", cur);
    delay(2);
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("current: %d\n", status.current_value);

    motor.write_mit(0, vel, pos, kp, kd);

    // control velocity
    kp = 0;
    kd = 13.1;
    motor.write_mit(500, 600, pos, kp, kd);
    printf("target_velocity: %d\n", 600);
    delay(2);
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("velocity: %d\n", status.velocity_value);
    motor.write_mit(0, 0, pos, kp, kd);

    // control position
    kp = 0.09;
    kd = 13.1;
    motor.write_mit(0, 0, 0, kp, kd);
    motor.write_mit(0, 0, 10000, kp, kd);
    printf("target_position: %d\n", 10000);
    delay(2);
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("position: %d\n", status.actuator_position_value);

    // header_types
    // 0xb1(list): Receive the content of the first status packet
    // 0xb2(list): Receive the content of the second status packet
    // 0xb3(list): Receive the content of the third status packet
    vector<uint8_t> header_types = {0xb1, 0xb2, 0xb3};
    status = motor.get_motor_status({status_packet1,status_packet2,status_packet3});
    printf("\nvoltage: %d pwm: %d current: %d velocity_value: %d\n", 
           status.voltage_value, status.pwm_value, status.current_value, status.velocity_value);
    printf("motor_position: %d actuator_position: %d\n", 
           status.motor_position_value, status.actuator_position_value);
    printf("mos_temp: %d motor_temp: %d g431_temp: %d warning_type: %d error_type: %d\n", 
           status.mos_temp, status.mt_temp, status.g431_temp, status.warning_type, status.error_type);
}
